#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
 
# 
# Transform executable FM statements.
#

sub xformStatements
{
    local($_);
    local($indent);
    local($flavor);
    local($line, $tree, $tree_type);

#    print "In xformStatements\n";
    
    for (@RawCode)
    {
	if (/^DECL:(\d+)/)
	{
	    &error($1, "Declaration amid executable statements. Is there a syntax error in a type declaration?");
	    next;
	}
	($line, $tree) = split(/\002/, $_, 2);
	if (!$tree )
	{
	    &addBodyCode($_);
	    next;
	}

	if (!($tree =~ s/^EXEC://))
	{
	    &error("Huh? xformStatements got invalid line '$_'\n");
	}

	$tree_type = &tree_type_name($tree);

	if ($tree_type eq "channel")
	{
	    &xformChannel($tree);
	}
	elsif ($tree_type eq "channel_close")
	{
	    &xformChannelClose($tree);
	}
	elsif ($tree_type eq "merger")
	{
	    &xformMerger($tree);
	}
	elsif ($tree_type eq "send")
	{
	    &xformSend($tree);
	}
	elsif ($tree_type eq "recv")
	{
	    &xformRecv($tree);
	}
	elsif ($tree_type eq "probe")
	{
	    &xformProbe($tree);
	}
	elsif ($tree_type eq "move_port")
	{
	    &xformMoveport($tree);
	}
	else
	{
	    &addBodyCode($_);
	}
    }

    # I'm not doing this check since to do it right we have to check
    # for ports that were sent to the thread in a message.

    # &checkCoupledOutputPorts;

    @RawCode = ();
}

sub xformSend
{
    local($tree) = @_;
    local($port, @args, @argstrs, $port_type);
    local(@contents, $elt, $type_name, $my_port_name, $elt_id, $dummy_dims);
    local(@args, $i, $arg, $type, $var);
    local($atype, $vtype, $port_idx);
    local(@callArgs, $callArgs, $port_tree);
    local(@subscr, $portstr);
    local(%specs, $origtype);

    &get_specs(&tree_kid($tree, 0), *specs);

    $port_tree = $specs{"port"};

    ($port) = &get_port_name($port_tree);
    $portstr = &array_element_name_to_str($port_tree);
    @args = &expand_tree_list(&tree_kid($tree, 1));

    if (($origtype = &vardict'getOriginalType($port)) ne 'out_port')
    {
	&error($tree,
               "SEND must take an out_port (port was $port, type was $origtype)");
	return;
    }
#    print "port '$port' origtype='$origtype'\n";

    $port_type = &vardict'get($port, 'port_ptr');

    if ($port_type =~ /^!port_(\d+)$/)
    {
	$port_idx = $1;
    }
    else
    {
	&error($tree, "No port_ptr for port $port");
	return;
    }

    grep(push(@argstrs, &expr_to_string($_)), @args);

#    print "Got send '$port' '@argstrs'\n";

    @contents = split(/\+/, &vardict'get($port, 'content_cache'));
    @contents || &error($tree, "No content_cache for $port");

    @contents == @args ||
	&error($tree, "SEND: argument mismatch for port $port");

    for ($i = 0; $i < @contents; $i++)
    {
	$var = '';
	
	$elt = $contents[$i];
	$arg = $args[$i];
	$argstr = $argstrs[$i];
	
	($type, $my_port_name, $elt_id,  $dummy_dims) =
	    split(/@/, $elt);

	#
	# Do type checking if we can
	#
	$atype = &tree_type_name($arg);
	if ($atype eq "string" || $atype eq "array_element_name")
	{
	    ($var) = &get_port_name($arg);
	    if (&vardict'get($var, 'what') eq "var")
	    {
		#
		# If it's a port, the port type is in OriginalType
		#
		$vtype = &vardict'getOriginalType($var);
		$vtype = &vardict'getType($var) if $vtype eq '';
		if ($type ne $vtype)
		{
		    local($") = ", "; # ")
		    # Just a warning for now...
		    &warning($tree,
			     "Type mismatch in SEND($port, @argstrs), arg " . ($i+2),
			     "(arg type is $vtype, port type is $type)");
		}
	    }
	}
	
	push(@callArgs, $argstr);
	if ($dummy_dims)
        {
	    local(@ndims, @dims, $size);
	    
	    @dims = split(/\|/, $dummy_dims);
#	    print "Dims is <", join(":", @dims), ">\n";
	    @ndims = ();
	    for (@dims)
	    {
		if ($_ =~ /^A_(\d+)$/)
		{
		    $_ = $argstrs[$1];
		}
		push(@ndims, "($_)");
	    }

#	    print "ndims is <", join(":", @ndims), ">\n";
	    $size = join(" * ", @ndims);
	    push(@callArgs, $size);
        }
    }

    local($iostat, $err_lab, $have_iostat);
    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};

    local($debug_suffix, $debug_vars);

    if ($opt_g)
    {
	local($file, $line);
	
	$debug_suffix = "d";
	&tree_line_info($tree, $file, $line);
	$debug_vars = ", '$file', $line";
    }
    else
    {
	$debug_suffix = '';
	$debug_vars = '';
    }

    if ($err_lab)
    {
	$err_lab = &tree_value($err_lab);
    }

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }

    $callArgs = join(", ", @callArgs);
    &addBodyCode("call fmu_s${FileBaseNoSpecials}$port_idx$debug_suffix($portstr, $callArgs, $err_var, $have_iostat $debug_vars)");
    if ($err_lab)
    {
	&addBodyCode("if ($err_var .gt. 0) goto $err_lab");
    }
}

sub xformRecv
{
    local($tree) = @_;
    local($port, @args, @argstrs, $port_type);
    local(@contents, $elt, $type_name, $my_port_name, $elt_id, $dummy_dims);
    local(@args, $i, $arg, $type, $var);
    local($atype, $vtype, $port_idx);
    local(@callArgs, $callArgs);
    local($port_tree, $portstr);
    local(%specs);

    &get_specs(&tree_kid($tree, 0), *specs);

    $port_tree = $specs{"port"};

    ($port) = &get_port_name($port_tree);
    $portstr = &array_element_name_to_str($port_tree);

    @args = &expand_tree_list(&tree_kid($tree, 1));
    grep(push(@argstrs, &expr_to_string($_)), @args);

    $port_type = &vardict'get($port, 'port_ptr');

    if (&vardict'getOriginalType($port) ne 'in_port')
    {
	&error($tree, "RECV must take an in_port (port was $port)");
	return;
    }

    if ($port_type =~ /^!port_(\d+)$/)
    {
	$port_idx = $1;
    }
    else
    {
	&error($tree, "No port_ptr for port $port");
	return;
    }


#    print "Got recv '$port' '@argstrs'\n";

    @contents = split(/\+/, &vardict'get($port, 'content_cache'));
    if (!@contents)
    {
	&error($tree, "No content_cache for $port");
	return;
    }

    if (@contents != @args)
    {
	&error($tree, "RECEIVE: argument mismatch for port $port");
	return;
    }

    for ($i = 0; $i < @contents; $i++)
    {
	$var = '';
	
	$elt = $contents[$i];
	$arg = $args[$i];
	$argstr = $argstrs[$i];
	
	($type, $my_port_name, $elt_id,  $dummy_dims) =
	    split(/@/, $elt);

	#
	# Do type checking. It's an error if we can't ... but the
	# grammar should enforce this.
	#
	$atype = &tree_type_name($arg);
	if ($atype eq "string" || $atype eq "array_element_name")
	{
	    ($var) = &get_port_name($arg);
	    if (&vardict'get($var, 'what') eq "var")
	    {
		#
		# If it's a port, the port type is in OriginalType
		#
		$vtype = &vardict'getOriginalType($var);
		$vtype = &vardict'getType($var) if $vtype eq '';
		if ($type ne $vtype)
		{
		    local($") = ", "; # ")
		    # Just a warning for now...
		    &warning($tree,
			     "Type mismatch in RECEIVE($port, @argstrs), arg " . ($i+2),
			     "(arg type is $vtype, port type is $type)");
		}
	    }
	}
	else
	{
	    &error($tree, "Args to RECV must be variable references");
	}
    }

    local($debug_suffix, $debug_vars);

    if ($opt_g)
    {
	local($file, $line);
	
	$debug_suffix = "d";
	&tree_line_info($tree, $file, $line);
	$debug_vars = ", '$file', $line";
    }
    else
    {
	$debug_suffix = '';
	$debug_vars = '';
    }


    local($iostat, $err_lab, $end_lab, $have_iostat);
    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};
    $end_lab = $specs{"end"};

    $err_lab = &tree_value($err_lab) if $err_lab;
    $end_lab = &tree_value($end_lab) if $end_lab;

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }

    $callArgs = join(", ", @argstrs);

    &addBodyCode("call fmu_r${FileBaseNoSpecials}$port_idx$debug_suffix($portstr, $callArgs, $err_var, $have_iostat $debug_vars)");
    if ($err_lab)
    {
	&addBodyCode("if ($err_var .gt. 0) goto $err_lab");
    }
    if ($end_lab)
    {
	&addBodyCode("if ($err_var .lt. 0) goto $end_lab");
    }
}

sub xformChannel
{
    local($tree) = @_;
    local(%specs, $i, $kid);
    local($pi_name, @pi_expr, $po_name, @po_expr, $pi_expr, $po_expr);
    local($err_var, $iostat, $err_lab, $have_iostat);
    
#    print "Got xformChannel:\n";

    $tree = &tree_kid($tree, 0);

    &get_specs($tree, *specs);

    $pi = $specs{"in"};
    $po = $specs{"out"};
    
    ($pi_name,@pi_expr) = &get_port_name($pi);
    ($po_name,@po_expr) = &get_port_name($po);


    $pi = $pi_name;
    $po = $po_name;

    if (@pi_expr)
    {
	grep($_ = &expr_to_string($_), @pi_expr);
	$pi.= '(' . join(",", @pi_expr) . ')';
    }    

    if (@po_expr)
    {
	grep($_ = &expr_to_string($_), @po_expr);
	$po.= '(' . join(",", @po_expr) . ')';
    }    

#    print "Got '$pi_name' '@pi_expr' '$po_name' '@po_expr'\n";

    if (&vardict'getOriginalType($pi_name) ne 'in_port')
    {
	&error($tree,
	       "Error: argument 1 '$pi_name' of CHANNEL statement must be an in_port");
    }
    
    if (&vardict'getOriginalType($po_name) ne 'out_port')
    {
	&error($tree,
	       "Error: argument 2 '$po_name' of CHANNEL statement must be an out_port");
    }

    if (&isArg($pi_name))
    {
	&error($tree,
	       "Error: argument 1 '$pi_name' of CHANNEL statement cannot be a subprogram argument");
    }

    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }
    &addBodyCode("call fm_channel($pi, $po, $err_var, $have_iostat)");
    if ($err_lab)
    {
	$err_lab = &tree_value($err_lab);
	&addBodyCode("if ($err_var .gt. 0) goto $err_lab");
    }
}

sub xformChannelClose
{
    local($tree) = @_;
    local($pi, $po, %specs);
    local($po_name, @po_expr, $po_expr);

    &get_specs(&tree_kid($tree, 0), *specs);
    
    $po = $specs{"port"};
    
    ($po_name,@po_expr) = &get_port_name($po);

    $po = $po_name;

    if (@po_expr)
    {
	grep($_ = &expr_to_string($_), @po_expr);
	$po.= '(' . join(",", @po_expr) . ')';
    }    

#    print "Got '$pi_name' '@pi_expr' '$po_name' '@po_expr'\n";

    if (&vardict'getOriginalType($po_name) ne 'out_port')
    {
	&error($tree,
	       "argument 1 '$po_name' of ENDCHANNEL statement must be an out_port");
    }

    local($iostat, $err_lab, $have_iostat);
    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};

    if ($err_lab)
    {
	$err_lab = &tree_value($err_lab);
    }

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }

    &addBodyCode("call fm_close_port($po, $err_var, $have_iostat)");

    if ($err_lab)
    {
	&addBodyCode("if ($err_var .gt. 0) goto $err_lab");
    }
}

sub xformProbe
{
    local($tree) = @_;
    local($pi, %specs);
    local($pi_name, @pi_expr, $pi_expr);

    &get_specs(&tree_kid($tree, 0), *specs);
    
    $pi = $specs{"port"};
    
    ($pi_name,@pi_expr) = &get_port_name($pi);

    $pi = $pi_name;

    if (@pi_expr)
    {
	grep($_ = &expr_to_string($_), @pi_expr);
	$pi.= '(' . join(",", @pi_expr) . ')';
    }    

#    print "Got '$pi_name' '@pi_expr'\n";

    if (&vardict'getOriginalType($pi_name) ne 'in_port')
    {
	&error($tree,
	       "argument 1 '$pi_name' of PROBE statement must be an in_port");
    }

    local($iostat, $err_lab, $empty_var, $have_iostat);
    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};
    $empty_var = $specs{"empty"};

    #
    # 5/3/93 empty is a variable, not a label. Retaining the _lab
    # to minimize changes to the code at this point.
    #
    # 7/28/93 really making empty a variable.
    #

    $err_lab = &tree_value($err_lab) if $err_lab;
    $empty_var = &tree_value($empty_var) if $empty_var;

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }

    if (!$empty_var)
    {
	$empty_var = &declareTemp('integer');
    }
    
    &addBodyCode("call fm_probe($pi, $err_var, $have_iostat, $empty_var)");

    if ($err_lab)
    {
	&addBodyCode("if ($err_var .gt. 0) goto $err_lab");
    }
}

sub xformMoveport
{
    local($tree) = @_;
    local(%specs);
    local($to_port, $to_name, @to_expr, $to_expr);
    local($from_port, $from_name, @from_expr, $from_expr);
    local($to_type, $from_type);
    local($to_dim, $from_dim);
    local($decl_size, $dim_size);
    local($call);

    &get_specs(&tree_kid($tree, 0), *specs);
    
    $to_port = $specs{"to"};
    $from_port = $specs{"from"};
    
    ($to_name,@to_expr) = &get_port_name($to_port);
    ($from_name,@from_expr) = &get_port_name($from_port);

    $to = $to_name;
    $from = $from_name;

    if (@to_expr)
    {
	grep($_ = &expr_to_string($_), @to_expr);
	$to .= '(' . join(",", @to_expr) . ')';
    }    

    if (@from_expr)
    {
	grep($_ = &expr_to_string($_), @from_expr);
	$from .= '(' . join(",", @from_expr) . ')';
    }

    $to_dim = &vardict'getDimension($to_name);
    $from_dim = &vardict'getDimension($from_name);

#    print "Got '$to_name' '@to_expr' '$to_dim' tree=$tree\n";
#    print "Got '$from_name' '@from_expr' '$from_dim'\n";

    $to_type = &vardict'getOriginalType($to_name);
    $from_type = &vardict'getOriginalType($from_name);

#    print "to_type=$to_type from_type=$from_type\n";

    &error($tree, "Invalid argument $to of MOVEPORT (not a port)")
	if !&isPortType($to_type);
	
    &error($tree, "Invalid argument $from of MOVEPORT (not a port)")
	if !&isPortType($from_type);

    &error($tree, "Arguments of MOVEPORT(to=$to, from=$from) must be of the same type")
	if $to_type ne $from_type;

    if ($to_type eq 'in_port')
    {
	$call = "fm_move_inport";
    }
    elsif ($to_type eq 'out_port')
    {
	$call = "fm_move_outport";
    }
    else
    {
	return;
    }
	

    if (@to_expr)
    {
	@decl_dims = split(/\|/, $to_dim);
	$decl_size = @decl_dims;
	$use_size = @to_expr;
#	print "to: decl_size=$decl_size use_size=$use_size\n";
        if ($decl_size < $use_size)
        {
            &error($tree, "MOVEPORT argument $to has too many subscripts");
        }
        elsif ($decl_size > $use_size)
        {
            &error($tree, "MOVEPORT argument $to is not a scalar");
        }
    }
    else
    {
	&error($tree, "MOVEPORT argument $to is not a scalar")
	    if $to_dim ne '';
    }

    if (@from_expr)
    {
	@decl_dims = split(/\|/, $from_dim);
	$decl_size = @decl_dims;
	$use_size = @from_expr;
#	print "from: decl_size=$decl_size use_size=$use_size\n";
        if ($decl_size < $use_size)
        {
            &error($tree, "MOVEPORT argument $from has too many subscripts");
        }
        elsif ($decl_size > $use_size)
        {
            &error($tree, "MOVEPORT argument $from is not a scalar");
        }
    }
    else
    {
	&error($tree, "MOVEPORT argument $from is not a scalar")
	    if $from_dim ne '';
    }

    &addBodyCode("call $call($from, $to)");
}

sub xformMerger
{
    local($tree) = @_;
    local($pi, $po_list, $po, %specs );
    local($pi_name, @pi_expr, $po_name, @po_expr, $pi_expr, $po_expr);
    local($n_slots, $phase);
    
    &get_specs(&tree_kid($tree, 0), *specs, 1);

    $pi = $specs{"in"};
    $po_list = $specs{"out"};

    ($pi_name,@pi_expr) = &get_port_name($pi);

    $pi = $pi_name;

    if (@pi_expr)
    {
	grep($_ = &expr_to_string($_), @pi_expr);
	$pi.= '(' . join(",", @pi_expr) . ')';
    }
    
    if (&vardict'getOriginalType($pi_name) ne 'in_port')
    {
	&error($tree,
	       "argument 1 '$pi_name' of MERGER statement must be an in_port");
    }
    
    if (&isArg($pi_name))
    {
	&error($tree,
	       "argument 1 '$pi_name' of MERGE statement cannot be a subprogram argument");
    }

    local($iostat, $err_lab, $gen_error_block, $have_iostat);
    $iostat = $specs{"iostat"};
    $err_lab = $specs{"err"};

    if ($err_lab)
    {
	$err_lab = &tree_value($err_lab);
    }
    else
    {
	$err_lab = $NextTempLabel++;
	$gen_error_block = 1;
    }

    $have_iostat = $iostat ? 1 : 0;

    if ($iostat)
    {
	$err_var = &get_var_name_as_string($iostat);
    }
    else
    {
	$err_var = &declareTemp('integer');
    }

    local($n_slots, $slot_idx);

    $slot_idx = &declareTemp('integer');

    for $phase ("init", "gen")
    {
	$n_slots = "0" if $phase eq "init";
	
#	print "Phase is <$phase>\n";
	if ($phase eq "gen")
	{
#	    print "Gen: got nslots=<$n_slots>\n";
	    &addBodyCode("call fm_merger_setup($n_slots, $pi, $slot_idx, $err_var, $have_iostat)");
	}
	for $po (&expand_tree_list($po_list))
	{
#	    print "po is <$po>\n";

	    if (&tree_type_name($po) eq "do_list")
	    {
		local($targ_list, $iter_var, $range, $lo, $hi, $step, $nreps);
		local(@vars, $n_vars, $var);

		$targ_list = &tree_kid($po, 0);

		for (&expand_tree_list($targ_list))
		{
		    if (&tree_type_name($_) eq "do_list")
		    {
			&fatal($_, "Nested implied do loops are not yet implemented");
		    }
		    push(@vars, &expr_to_string($_));
		}

		$n_vars = @vars;
	    
		$iter_var = &expr_to_string(&tree_kid($po, 1));
		$range = &tree_kid($po, 2);
		
#	    &dump_parse_tree($po);
#	    &dump_parse_tree($targ_list);
#		&dump_parse_tree($range);
		
		$lo = &expr_to_string(&tree_kid($range, 0));
#		$lo = &tree_value(&tree_kid($range, 0));
		if (&tree_num_kids($range) >= 2)
		{
		    $hi = &expr_to_string(&tree_kid($range, 1));
#		    $hi = &tree_value(&tree_kid($range, 1));
		    if (&tree_num_kids($range) >= 3)
		    {
			$step = &expr_to_string(&tree_kid($range, 2));
#			$step = &tree_value(&tree_kid($range, 2));
		    }
		    else
		    {
			$step = 1;
		    }
		}
		else
		{
		    $hi = $lo;
		    $step = 1;
		}
		
#		print "Have do-list var=<$var> lo=<$lo> hi=<$hi> step=<$step>\n";
#		print "Vars=<", join(":", @vars), ">\n";


		if ($phase eq "init")
		{
		    $n_slots = "$n_slots + $n_vars * ( ( $hi - $lo + $step) / $step)";
		}
		else
		{
		    &addBodyCode("do $iter_var = $lo, $hi, $step");
		    for $var (@vars)
		    {
			&addBodyCode("   call fm_merger_add($var, $pi, $slot_idx, $err_var, $have_iostat)");
		    }
		    &addBodyCode("enddo");
		}
	    }
	    else
	    {
		($po_name,@po_expr) = &get_port_name($po);
		
		$po = $po_name;
		
		if (@po_expr)
		{
		    grep($_ = &expr_to_string($_), @po_expr);
		    $po.= '(' . join(",", @po_expr) . ')';
		}    
	    
#		print "Got '$pi_name' '@pi_expr' '$po_name' '@po_expr'\n";
		
		if (&vardict'getOriginalType($po_name) ne 'out_port')
		{
		    &error($tree, "argument 2 '$po_name' of MERGE statement must be an out_port");
		}

		if ($phase eq "init")
		{
		    $n_slots = "$n_slots + 1";
		}
		else
		{
		    &addBodyCode("call fm_merger_add($po, $pi, $slot_idx, $err_var, $have_iostat)");
		}
	    }
	}
    }

    &addBodyCode("call fm_merger_done($pi, $slot_idx, $err_var, $have_iostat)");
    &addBodyCode("if ($err_var .lt. 0) goto $err_lab");

    if ($gen_error_block)
    {
	&addBodyCode("$err_lab continue");
    }
}

	    
1;
