/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#ifdef FM_HAS_INLINE
#define INLINE inline
#else
#define INLINE
#endif

/*
 * Macros for the standard methods for dealing with incoming messages.
 */
#define ProcessSendMessage(type, buf) \
	_fm_process_enqueue_message(FM_CHANNEL_OF_TYPE(type), buf, FALSE);

#define ProcessForwardMessage(type, buf) \
	_fm_process_enqueue_message(FM_CHANNEL_OF_TYPE(type), buf, TRUE);


#define ProcessLastForwardMessage(type, buf) \
	_fm_do_last_forward_message(type, buf); 

#define ProcessCloseMessage(type, buf) \
	{ \
	     int inport_idx; \
	     yoke_buffer_t rbuf; \
	     conn_t reply_conn; \
	     yoke_get_int(buf, &inport_idx, 1); \
	     yoke_get_conn(buf, &reply_conn); \
	     rbuf = yoke_allocate_buffer(reply_conn, 0, 0); \
	     yoke_establish(reply_conn); \
	     yoke_send(reply_conn, rbuf, FM_MTYPE_CLOSE_ACK); \
	     yoke_release(reply_conn); \
	     _fm_process_close_channel(FM_CHANNEL_OF_TYPE(type), inport_idx); \
	 }

#define ProcessPermReq(type, buf) \
	_fm_process_permission_req(buf, TRUE);
     
#define ProcessPermReqNoQueueProcessing(type, buf) \
	_fm_process_permission_req(buf, FALSE);
     
#define ProcessPermReqWithQueueFlag(type, buf, flag) \
	_fm_process_permission_req(buf, flag);
     

/*
 * Code for bracketing receive loops. Each receive loop must include
 * a case for FM_MTYPE_INPORT_MIGRATE and FM_MTYPE_OUTPORT_MIGRATE,
 * to be handled by the HandleInportMigrate and HandleOutportMigrate
 * macros. The StartReceiveLoop macro is invoked before the
 * loop begins, and EndReceiveLoop is invoked after the loop is finished.
 *
 * The intent here is to keep the notification signals out of the
 * hair of the send/recv code.
 *
 */

#ifdef DEBUG
#define HCA_DEBUG(sender) \
	{if (PortDebug(2)) \
	    yoke_printf("Got close ack from %s\n",\
			yoke_conn_as_string(sender));}
#else
#define HCA_DEBUG(sender)
#endif

#define HandleCloseAck(sender, buf) \
	{if (_fm_n_close_acks_needed <= 0) \
	    yoke_fatal("Got close ack from %s when we didn't need one at %s:%d\n", \
		       yoke_conn_as_string(sender), \
		       __FILE__, __LINE__); \
	HCA_DEBUG(sender); \
	yoke_free_buffer(buf); \
	_fm_n_close_acks_needed--;}

#define HandlePortSend(sender, buf) \
	_fm_handle_port_send(sender, buf, FALSE);

#define EnqueuePortSend(sender, buf) \
	_fm_handle_port_send(sender, buf, TRUE);

#define HandleInportMigrate(sender, buf) \
	_fm_do_inport_migration(sender, buf, TRUE);

#define HandleOutportMigrate(sender, buf) \
	_fm_do_outport_migration(sender, buf, TRUE);

#define HandleProcessReturn(sender, buf) \
	_fm_handle_process_return_message(sender, buf);

#define StartSafeRegionGuts \
	yoke_block_notification();

#define EndSafeRegionGuts \
	yoke_unblock_notification();

#ifdef DEBUG

#define StartSafeRegion \
        {if (MigrateDebug(4)) \
	    yoke_printf("Entering receive loop at %s:%d\n", \
			__FILE__, __LINE__); \
	StartSafeRegionGuts;}

#define EndSafeRegion \
        {if (MigrateDebug(4)) \
	    yoke_printf("Exiting receive loop at %s:%d\n", \
			__FILE__, __LINE__); \
	EndSafeRegionGuts;}

#else /* DEBUG */

#define StartSafeRegion StartSafeRegionGuts

#define EndSafeRegion EndSafeRegionGuts

#endif /* DEBUG */

/*
 * Message types.
 *
 * There are two flavors of messages types: those with a channel
 * id as part of the type and those without. The channel id is
 * encoded as the value of the id shifted left FM_N_TAG_BITS bits
 * and or'd with the message tag 1 thru 7. Message tag 0 is
 * reserved to designate message types which need no channel
 * id.
 */

#define FM_N_TAG_BITS		3
#define FM_TAG_BITS_MASK	0x7
#define FM_SHIFT_FOR_TYPE(val)	((val) << (FM_N_TAG_BITS))

#define FM_MTYPE_NODE_INITIALIZATION	FM_SHIFT_FOR_TYPE(10)
#define FM_MTYPE_MAIN_RUN		FM_SHIFT_FOR_TYPE(11)

#define FM_CREATE_PROCESS		FM_SHIFT_FOR_TYPE(12)
#define FM_CREATE_PROCESS_RETURN	FM_SHIFT_FOR_TYPE(13)

#define FM_START_PROCESS		FM_SHIFT_FOR_TYPE(14)

#define FM_PROCESS_RETURN		FM_SHIFT_FOR_TYPE(15)

#define FM_SIGNAL_CHILD			FM_SHIFT_FOR_TYPE(16)

#define FM_CHANNEL_SEND			FM_SHIFT_FOR_TYPE(17)


#define FM_MTYPE_SHUTDOWN_REQUEST	FM_SHIFT_FOR_TYPE(18)
#define FM_MTYPE_MGR_SHUTDOWN_1		FM_SHIFT_FOR_TYPE(19)
#define FM_MTYPE_MGR_SHUTDOWN_1_ACK	FM_SHIFT_FOR_TYPE(20)
#define FM_MTYPE_MGR_SHUTDOWN_2		FM_SHIFT_FOR_TYPE(21)
#define FM_MTYPE_ABEND			FM_SHIFT_FOR_TYPE(22)

#define FM_MTYPE_YOUR_MANAGER		FM_SHIFT_FOR_TYPE(23)
#define FM_MTYPE_YOUR_MANAGER_ACK	FM_SHIFT_FOR_TYPE(24)

#define FM_MTYPE_MANAGER_LIST_REQ	FM_SHIFT_FOR_TYPE(25)
#define FM_MTYPE_MANAGER_LIST_REQ_REPLY	FM_SHIFT_FOR_TYPE(26)

#define FM_MANAGER_QUIT_ACK_ACK		FM_SHIFT_FOR_TYPE(27)

#define FM_MTYPE_INPORT_MIGRATE		FM_SHIFT_FOR_TYPE(28)
#define FM_MTYPE_INPORT_MGR_MIGRATE	FM_SHIFT_FOR_TYPE(29)
#define FM_MTYPE_INPORT_MIGRATE_ACK	FM_SHIFT_FOR_TYPE(30)

#define FM_MTYPE_OUTPORT_MIGRATE	FM_SHIFT_FOR_TYPE(31)
#define FM_MTYPE_OUTPORT_MGR_MIGRATE	FM_SHIFT_FOR_TYPE(32)
#define FM_MTYPE_OUTPORT_MIGRATE_ACK	FM_SHIFT_FOR_TYPE(33)

#define FM_MTYPE_MANAGER_READY		FM_SHIFT_FOR_TYPE(34)

#define FM_MTYPE_CHILD_EXITING		FM_SHIFT_FOR_TYPE(35)
#define FM_MTYPE_CHILD_EXITING_ACK	FM_SHIFT_FOR_TYPE(36)

#define FM_MTYPE_SEND_PORTS		FM_SHIFT_FOR_TYPE(37)
#define FM_MTYPE_SEND_PORTS_REPLY	FM_SHIFT_FOR_TYPE(38)

#define FM_MTYPE_CLOSE_ACK		FM_SHIFT_FOR_TYPE(39)

#define FM_MTYPE_DO_SHUTDOWN		FM_SHIFT_FOR_TYPE(40)

#define FM_MTYPE_OUTPORT_PERM_REQ	FM_SHIFT_FOR_TYPE(41)
#define FM_MTYPE_OUTPORT_PERM_GRANTED	FM_SHIFT_FOR_TYPE(42)

#define FM_MTYPE_INPORT_MIGRATE_NO_ACK	FM_SHIFT_FOR_TYPE(43)
#define FM_MTYPE_OUTPORT_MIGRATE_NO_ACK	FM_SHIFT_FOR_TYPE(44)

#define FM_SIGNAL_CHILD_REPLY		FM_SHIFT_FOR_TYPE(45)
#define FM_MTYPE_SHUTDOWN_REQUEST_ACK	FM_SHIFT_FOR_TYPE(46)

#define FM_SEND_TYPE(chan)	(FM_SHIFT_FOR_TYPE(chan) | FM_FLAVOR_SEND)
#define FM_CLOSE_TYPE(chan)	(FM_SHIFT_FOR_TYPE(chan) | FM_FLAVOR_CLOSE)
#define FM_FWD_BUF_TYPE(chan)	(FM_SHIFT_FOR_TYPE(chan) | FM_FLAVOR_FWD_BUF)
#define FM_FWD_LAST_BUF_TYPE(chan)	(FM_SHIFT_FOR_TYPE(chan) | FM_FLAVOR_FWD_LAST_BUF)

#define FM_TYPE_FLAVOR(type)		(type & FM_TAG_BITS_MASK)

#define FM_FLAVOR_CONTROL		0x0
#define FM_FLAVOR_SEND			0x1
#define FM_FLAVOR_CLOSE			0x2
#define FM_FLAVOR_FWD_BUF		0x3
#define FM_FLAVOR_FWD_LAST_BUF		0x4

#define FM_IS_CONTROL(type)	(FM_TYPE_FLAVOR(type) == FM_FLAVOR_CONTROL)
#define FM_IS_SEND(type)	(FM_TYPE_FLAVOR(type) == FM_FLAVOR_SEND)
#define FM_IS_CLOSE(type)	(FM_TYPE_FLAVOR(type) == FM_FLAVOR_CLOSE)
#define FM_IS_FWD_BUF(type)	(FM_TYPE_FLAVOR(type) == FM_FLAVOR_FWD_BUF)
#define FM_IS_FWD_LAST_BUF(type) (FM_TYPE_FLAVOR(type) == FM_FLAVOR_FWD_LAST_BUF)

#define FM_CHANNEL_OF_TYPE(type)	(type >> FM_N_TAG_BITS)

#define FM_HANDLER_ID_START		1
#define FM_MAX_REMOTE_START_HANDLERS	1000

#define FM_IOSTAT_EOF			-1
#define FM_IOSTAT_OK			0
#define FM_IOSTAT_ERROR			1
#define FM_IOSTAT_INPORT_UNCONNECTED		2
#define FM_IOSTAT_OUTPORT_UNCONNECTED		3
#define FM_IOSTAT_INPORT_ALREADY_CONNECTED	4
#define FM_IOSTAT_OUTPORT_ALREADY_CONNECTED	5
#define FM_IOSTAT_CLOSE_TWICE			6

#ifdef DEBUG
#define SetIOSTAT(error) \
	_fm_set_iostat_debug(iostat, (int) *have_iostat, error, __FILE__, __LINE__)
#define SetIOSTATWithLine(error, file, line) \
	_fm_set_iostat_debug(iostat, (int) *have_iostat, error, file, line)
#else
#define SetIOSTAT(error) \
	{if (error == FM_IOSTAT_OK || error == FM_IOSTAT_EOF) \
	    *iostat = error; \
	else \
	    _fm_set_iostat(iostat, (int) *have_iostat, error);}
#endif


#ifdef DEBUG
#define FMDebug(Level) (_fm_debug_level >= (Level))
#define PortDebug(Level) (_fm_port_debug_level >= (Level))
#define MigrateDebug(Level) (_fm_migrate_debug_level >= (Level))
#define ProcessDebug(Level) (_fm_process_debug_level >= (Level))

#define Assert(condition) YokeAssert(condition);

#define PortName(port) ((port)->name)

#else /* DEBUG */

#define Assert(condition)

#define PortName(port) "<unknown>"

#endif /* DEBUG */

#ifdef MALLOC_DEBUG
#undef malloc
#undef free
#define malloc(size) yoke_debug_malloc(size, __FILE__, __LINE__)
#define free(ptr) yoke_debug_free(ptr, __FILE__, __LINE__)

#endif /* MALLOC_DEBUG */

#ifdef __STDC__
#define MallocCheck(func, var, type) \
    { if ((var) == (type) NULL) \
	yoke_fatal_malloc_5(#func, #type, #var, __FILE__, __LINE__); }

#else
#define MallocCheck(func, var, type) \
    { if ((var) == (type) NULL) \
	yoke_fatal_malloc_5("func", "type", "var", __FILE__, __LINE__); }
#endif

#ifdef DEBUG

#ifdef MALLOC_DEBUG
#define YOKE_MALLOC_CHECK \
	{if (yoke_malloc_debug_level() >= 3) yoke_malloc_check(__FILE__, __LINE__);}

#ifdef NeXT
#define FM_MALLOC_CHECK if (yoke_malloc_debug_level() >= 1) NXMallocCheck();
#endif /* NeXT */
#ifdef sun
#define FM_MALLOC_CHECK \
	{if ((yoke_malloc_debug_level() >= 1) && (malloc_verify() == 0)) \
	yoke_fatal("malloc_verify(): failed at %s:%d\n", \
		   __FILE__, __LINE__);}
#endif /* sun */
#endif /* MALLOC_DEBUG */

#ifndef FM_MALLOC_CHECK
#define FM_MALLOC_CHECK
#endif

#ifndef YOKE_MALLOC_CHECK
#define YOKE_MALLOC_CHECK
#endif

#ifdef __STDC__
#define FM_INPORT_MAGIC_CHECK(Func, Port)  { \
    YOKE_MALLOC_CHECK; \
    FM_MALLOC_CHECK; \
    if ((Port) == (fm_inport_t) NULL) \
	yoke_fatal("%s(): Unexpectedly received a null inport (%s:%d)\n", \
		   #Func, __FILE__, __LINE__); \
    if ((Port)->start_magic != FM_INPORT_START_MAGIC) \
	yoke_fatal("%s(): Received an invalid inport (start magic %x is wrong) (%s:%d)\n", \
		   #Func, (Port)->start_magic, __FILE__, __LINE__); \
    if ((Port)->end_magic != FM_INPORT_END_MAGIC) \
	yoke_fatal("%s(): Received an invalid inport (start magic %x is wrong) (%s:%d)\n", \
		   #Func, (Port)->start_magic, __FILE__, __LINE__); }
#else
#define FM_INPORT_MAGIC_CHECK(Func, Port)  { \
    YOKE_MALLOC_CHECK; \
    FM_MALLOC_CHECK; \
    if ((Port) == (fm_inport_t) NULL) \
	yoke_fatal("%s(): Unexpectedly received a null inport (%s:%d)\n", \
		   "Func", __FILE__, __LINE__); \
    if ((Port)->start_magic != FM_INPORT_START_MAGIC) \
	yoke_fatal("%s(): Received an invalid inport (start magic %x is wrong) (%s:%d)\n", \
		   "Func", (Port)->start_magic, __FILE__, __LINE__); \
    if ((Port)->end_magic != FM_INPORT_END_MAGIC) \
	yoke_fatal("%s(): Received an invalid inport (start magic %x is wrong) (%s:%d)\n", \
		   "Func", (Port)->start_magic, __FILE__, __LINE__); }
#endif

#ifdef __STDC__
#define FM_OUTPORT_MAGIC_CHECK(Func, Port) { \
    YOKE_MALLOC_CHECK; \
    FM_MALLOC_CHECK; \
    if ((Port) == (fm_outport_t) NULL) \
	yoke_fatal("%s(): Unexpectedly received a null outport (%s:%d)\n", \
		   #Func, __FILE__, __LINE__); \
    if ((Port)->start_magic != FM_OUTPORT_START_MAGIC) \
	yoke_fatal("%s(): Received an invalid outport (start magic %x is wrong) (%s:%d)\n", \
		   #Func, (Port)->start_magic, __FILE__, __LINE__); \
    if ((Port)->end_magic != FM_OUTPORT_END_MAGIC) \
	yoke_fatal("%s(): Received an invalid outport (start magic %x is wrong) (%s:%d)\n", \
		   #Func, (Port)->start_magic, __FILE__, __LINE__); }
#else
#define FM_OUTPORT_MAGIC_CHECK(Func, Port) { \
    YOKE_MALLOC_CHECK; \
    FM_MALLOC_CHECK; \
    if ((Port) == (fm_outport_t) NULL) \
	yoke_fatal("%s(): Unexpectedly received a null outport (%s:%d)\n", \
		   "Func", __FILE__, __LINE__); \
    if ((Port)->start_magic != FM_OUTPORT_START_MAGIC) \
	yoke_fatal("%s(): Received an invalid outport (start magic %x is wrong) (%s:%d)\n", \
		   "Func", (Port)->start_magic, __FILE__, __LINE__); \
    if ((Port)->end_magic != FM_OUTPORT_END_MAGIC) \
	yoke_fatal("%s(): Received an invalid outport (start magic %x is wrong) (%s:%d)\n", \
		   "Func", (Port)->start_magic, __FILE__, __LINE__); }
#endif

#define FM_INPORT_START_MAGIC	0x34343434
#define FM_INPORT_END_MAGIC	0x12121212

#define FM_OUTPORT_START_MAGIC	0x19191919
#define FM_OUTPORT_END_MAGIC	0x82828282

#else /* DEBUG */

#define FM_INPORT_MAGIC_CHECK(Func, Port)
    
#define FM_OUTPORT_MAGIC_CHECK(Func, Port)

#endif /* DEBUG */
    

#define MAX_ABEND_MSG_SIZE 1024

