/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

static fm_remote_start_func_registry_t handlers[FM_MAX_REMOTE_START_HANDLERS];

void _fm_register_handlers()
{
    extern fm_remote_start_func_registry_t *_fm_registry_tables[];
    fm_remote_start_func_registry_t **tblp, *tbl;

    bzero(handlers, sizeof(handlers));

    for (tblp = _fm_registry_tables; *tblp != NULL; tblp++)
    {
	for (tbl = *tblp; tbl->id != NULL; tbl++)
	{
#ifdef DEBUG
	    if (FMDebug(3))
		yoke_printf("Registering id=%d func=%x name=%s\n",
		   *(tbl->id), tbl->func, tbl->name);
#endif

	    handlers[*(tbl->id)] = *tbl;
	}
    }
}

fm_remote_start_func_t
    _fm_lookup_remote_start(fm_remote_start_func_id_t procedure_ident)
{
    if (handlers[procedure_ident].func == NULL)
	yoke_fatal("Cannot find procedure for ident %d\n", procedure_ident);

#ifdef DEBUG
    if (FMDebug(3))
	yoke_printf("Found procedure %x named %s for ident %d\n",
		    handlers[procedure_ident].func,
		    handlers[procedure_ident].name, procedure_ident);
#endif
    return handlers[procedure_ident].func;
}

char *
    _fm_lookup_remote_start_name(fm_remote_start_func_id_t procedure_ident)
{
    if (handlers[procedure_ident].func == NULL)
	return "<unknown procedure>";

    return handlers[procedure_ident].name;
}

char *
    _fm_lookup_process(fm_remote_start_func_id_t procedure_ident)
{
    if (handlers[procedure_ident].func == NULL)
	return "<unknown procedure>";

    return handlers[procedure_ident].process;
}

