/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include <stdio.h>
#include <string.h>

#include "yoke_internal.h"

#ifdef NeXT
#include <libc.h>
#endif

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


/*
 * yoke_silent_fatal()
 *
 * Fatal error out without printing any messages.
 */
void yoke_silent_fatal()
{
    yoke_proto_t proto;

    /* Call each of the proto's fatal routine */
    for (proto = _y_my_protos; proto; proto = proto->next)
    {
	(proto->funcs->fatal)();
    }

    /* Call the machine dependent fatal routine.  This should not return */
    _y_md_fatal(1);
} /* yoke_silent_fatal() */


/*
 * yoke_fatal()
 */
#ifdef __STDC__
void yoke_fatal(char *msg, ...)
#else
void yoke_fatal(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    char outbuf[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: Fatal error: %s", host, msg);

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vsprintf(outbuf, fmt, ap);
    va_end(ap);

    write(1, outbuf, strlen(outbuf));

    fflush(_y_stdout);

    if (_y_pause_on_fatal)
    {
	yoke_printf("Process %d pausing on fatal error\n", getpid());
	_y_pausing_for_fatal = TRUE;
	yoke_pause();
    }

    yoke_silent_fatal();
    
} /* yoke_fatal() */


/*
 * yoke_error()
 */
#ifdef __STDC__
void yoke_error(char *msg, ...)
#else
void yoke_error(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: Error: %s", host, msg);

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);
} /* yoke_error() */


/*
 * yoke_warning()
 */
#ifdef __STDC__
void yoke_warning(char *msg, ...)
#else
void yoke_warning(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: Warning: %s", host, msg);

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);
} /* yoke_warning() */


/*
 * yoke_notice()
 */
#ifdef __STDC__
void yoke_notice(char *msg, ...)
#else
void yoke_notice(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: Notice: %s", host, msg);

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);
} /* yoke_notice() */


/*
 * yoke_printf()
 */
#ifdef __STDC__
void yoke_printf(char *msg, ...)
#else
void yoke_printf(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: %s", host, msg);

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);
} /* yoke_printf() */


/*
 * yoke_perror()
 */
#ifdef __STDC__
void yoke_perror(char *msg, ...)
#else
void yoke_perror(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: %s: %s\n", host, msg, _y_md_system_error_string());

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);
} /* yoke_perror() */


#ifdef __STDC__
void yoke_fatal_perror(char *msg, ...)
#else
void yoke_fatal_perror(msg, va_alist)
char *msg;
va_dcl
#endif
{
    char fmt[1024];
    va_list ap;
    char *host;
    
    _y_md_get_short_host_id(&host);

    sprintf(fmt, "%15s: Fatal error: %s: %s\n", host, msg, _y_md_system_error_string());

#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vfprintf(_y_stdout, fmt, ap);
    fflush(_y_stdout);
    va_end(ap);

    if (_y_pause_on_fatal)
    {
	yoke_printf("Process %d pausing on fatal error\n", getpid());
	_y_pausing_for_fatal = TRUE;
	yoke_pause();
    }

    _y_md_fatal(1);
}

/*
 * yoke_assert_sprintf()
 *
 * This is used by the YokeAssert2() macro...
 */
#ifdef __STDC__
char *yoke_assert_sprintf(char *msg, ...)
#else
char *yoke_assert_sprintf(msg, va_alist)
char *msg;
va_dcl
#endif
{
    static char assert_sprintf_buf[1024];
    va_list ap;
    
#ifdef __STDC__
    va_start(ap, msg);
#else
    va_start(ap);
#endif
    vsprintf(assert_sprintf_buf, msg, ap);
    va_end(ap);

    return (assert_sprintf_buf);
} /* yoke_assert_sprintf() */


/*
 * yoke_fatal_malloc()
 *
 * Fatal out due to a failed malloc.
 *
 * Use write() instead of fprintf() for malloc errors, because if the
 * error was caused by a failed malloc(), then fprintf() will
 * segmentation fault -- evidently, it uses malloc().
 */
void yoke_fatal_malloc_5(char *func, char *type, char *var, char *file,
			 int line)
{
    static char err_buf[2048];
    char num[20], *s;
    int n;

#define MALLOC_ERROR_MSG "Fatal error: Failed malloc\n"

    s = num + sizeof(num) - 1;
    *s = '\0';
    for (n = line; n > 0;)
    {
	s--;
	*s = '0' + n % 10;
	n /= 10;
    }
    
    strcpy(err_buf, "Fatal error: ");
    strcat(err_buf, func);
    strcat(err_buf, ": malloc failed for ");
    strcat(err_buf, type);
    strcat(err_buf, " ");
    strcat(err_buf, var);
    strcat(err_buf, " in file ");
    strcat(err_buf, file);
    strcat(err_buf, " line ");
    strcat(err_buf, s);
    strcat(err_buf, "\n");

    write(2, err_buf, strlen(err_buf));
    
    yoke_silent_fatal();
} /* yoke_fatal_malloc_5() */

void yoke_fatal_malloc_6(char *func, int size,
			 char *type, char *var, char *file,
			 int line)
{
    static char err_buf[2048];
    char num[20], *s;
    char num2[20], *s2;
    int n;

    s = num + sizeof(num) - 1;
    *s = '\0';
    for (n = line; n > 0;)
    {
	s--;
	*s = '0' + n % 10;
	n /= 10;
    }
    
    s2 = num2 + sizeof(num2) - 1;
    *s2 = '\0';
    for (n = size; n > 0;)
    {
	s2--;
	*s2 = '0' + n % 10;
	n /= 10;
    }
    
    strcpy(err_buf, "Fatal error: ");
    strcat(err_buf, func);
    strcat(err_buf, ": malloc(");
    strcat(err_buf, s2);
    strcat(err_buf, ") failed for ");
    strcat(err_buf, type);
    strcat(err_buf, " ");
    strcat(err_buf, var);
    strcat(err_buf, " in file ");
    strcat(err_buf, file);
    strcat(err_buf, " line ");
    strcat(err_buf, s);
    strcat(err_buf, "\n");

    write(2, err_buf, strlen(err_buf));
    
    yoke_silent_fatal();
} /* yoke_fatal_malloc_6() */

